from fastapi import FastAPI, Body
from fastapi.responses import StreamingResponse
from typing import Union
from sdmx_mapping_utility import KeyRange, SDMXMappingUtility
import logging
FORMAT = '%(asctime)s: [%(levelname)s] %(message)s'
logging.basicConfig(format=FORMAT, level=logging.INFO)

app = FastAPI()

@app.get("/")
def info():
    return f"the partial key mapping service says hello"

@app.get("/map_withURN")
async def map_withURN(
                    source_file: str,
                    mapping: str,
                    registry: str,
                    return_source_dimensions: Union[bool, None] = False,
                    return_target_attributes: Union[bool, None] = False,
                    generate_nulls_for_fixed_targets: Union[bool, None] = False,
                    write_target_struct_info: Union[bool, None] = False,
                    drop_invalid_mappings: Union[bool, None] = False
                    ):
    """ Mapping with mapping provided as a structure-set in a registry \n
        e.g. mapping:  OECD.SDD.NAD/QNA_NAMAIN1_MAPPING/latest \n
        registry: http://sdmxri-dev-2.main.oecd.org:8080/FusionRegistry/ws/public/sdmxapi/rest 
    """
    print(f'map_withURN called with source_file: {source_file}, registry: {registry}, mapping: {mapping}')
    url = f'{registry}/structuremap/{mapping}/?detail=full&references=descendants'
    mu = SDMXMappingUtility.initialise_with_registry(registry_url=url, source_file=source_file)
    df = mu.generate_mappings_partial_keys(
            includeSourceColumns=return_source_dimensions,
            includeAttributesMeasures=return_target_attributes,
            nulledFixedTargets=generate_nulls_for_fixed_targets,
            writeTargetStructInfo=write_target_struct_info,
            dropInvalid=drop_invalid_mappings)
    
    
    return StreamingResponse(iter([df.to_csv(index=False)]), media_type="text/csv")
    
@app.get("/map_withFile")
async def map_withFile(
                    source_file: str,
                    mapping_file: str,
                    return_source_dimensions: Union[bool, None] = False,
                    return_target_attributes: Union[bool, None] = False,
                    generate_nulls_for_fixed_targets: Union[bool, None] = False,
                    write_target_struct_info: Union[bool, None] = False,
                    drop_invalid_mappings: Union[bool, None] = False, 
                    sid_based: Union[bool, None] = False
                    ):
    """ Mapping with mapping provided as a file with interrelated structures 
    """
    print( f'map_withFile called with source_file: {source_file}, mapping_file: {mapping_file}') 
    mu = SDMXMappingUtility.initialise_with_file(mapping_file=mapping_file, source_file=source_file, sid_based=sid_based )
    df = mu.generate_mappings_partial_keys(
            includeSourceColumns=return_source_dimensions,
            includeAttributesMeasures=return_target_attributes,
            nulledFixedTargets=generate_nulls_for_fixed_targets,
            writeTargetStructInfo=write_target_struct_info,
            dropInvalid=drop_invalid_mappings,
            sid_based=sid_based)
    

    return StreamingResponse(iter([df.to_csv(index=False)]), media_type="text/csv")

@app.post("/map_json_withURN")
async def map_json_withURN(
                    mapping: str,
                    registry: str,
                    json: KeyRange = Body(),
                    return_source_dimensions: Union[bool, None] = False,
                    return_target_attributes: Union[bool, None] = False,
                    generate_nulls_for_fixed_targets: Union[bool, None] = False,
                    write_target_struct_info: Union[bool, None] = False,
                    drop_invalid_mappings: Union[bool, None] = False, 
                    sid_based: Union[bool, None] = False
                    ):
    """ Mapping with mapping provided as a structure-set in a registry \n
        e.g. mapping:  OECD.SDD.NAD/QNA_NAMAIN1_MAPPING/latest \n
        registry: http://sdmxri-dev-2.main.oecd.org:8080/FusionRegistry/ws/public/sdmxapi/rest 
    """
    print(f'map_withURN called with source_file: {json}, registry: {registry}, mapping: {mapping}')
    url = f'{registry}/structuremap/{mapping}/?detail=full&references=descendants'
    mu = SDMXMappingUtility.initialise_with_jsonandregistry(registry_url=url, json=json, sid_based=sid_based)
    df = mu.generate_mappings_partial_keys(
            includeSourceColumns=return_source_dimensions,
            includeAttributesMeasures=return_target_attributes,
            nulledFixedTargets=generate_nulls_for_fixed_targets,
            writeTargetStructInfo=write_target_struct_info,
            dropInvalid=drop_invalid_mappings,
            sid_based=sid_based)
    
    
    return StreamingResponse(iter([df.to_csv(index=False)]), media_type="text/csv")


import pandas as pd
import collections as coll


class MappingUtility():
  '''An object to do operations on a key-to-key mapping table
  initialise with the mapping table, a list of source column headers and a list of target column headers
  '''

  def __init__(self, mappingDataFrame: pd.DataFrame, sourceList: list, targetList: list):
      self.mappingDafaFrame = mappingDataFrame
      self.sourceList = sourceList
      self.targetList = targetList
  
  @classmethod
  def from_csv(class_object, path: str, sourceList: list, targetList: list):
      return class_object(pd.read_csv(path), sourceList, targetList)


  @classmethod
  def from_dict(class_object, mappingDict: dict , sourceList: list, targetList: list):  
      return class_object(pd.DataFrame(mappingDict), sourceList, targetList)


  def generateGroupDimension(self, fromList: list):
      if not all(e in (self.sourceList + self.targetList) for e in fromList):
        raise ValueError('at least one of the names do not exist in the mapping dataframe')

      return self.mappingDafaFrame[fromList].apply(lambda x: pd.Series('.'.join(x), index=['.'.join(fromList)]), axis=1).iloc[:,0]


  def maxTargetVariation(self, sourceCodes, targetCodes):
      if not (isinstance(sourceCodes, list) and isinstance(targetCodes, list)):
        raise TypeError('inputs should be lists')
      
      #print(sourceCodes, targetCodes)
      uniquecombinations = { '.'.join((x,y)) : x for x, y in zip(sourceCodes, targetCodes)}
      countedsources = coll.Counter(uniquecombinations.values())
      return max(countedsources.values())

  def analyseOneToOneCompatibility(self):
      otoc = pd.DataFrame(index= self.sourceList, columns= self.targetList)
      for k in self.sourceList:
        for l in self.targetList:
          otoc.at[k,l] = self.maxTargetVariation( self.mappingDafaFrame[k].values.tolist(), self.mappingDafaFrame[l].values.tolist())    
     
      return otoc
  

  def getMinimumCompatibilitySet(self, targetDimension):
      sourceDims = self.generateGroupDimension(self.sourceList).values.tolist()
      targetDim = self.mappingDafaFrame[targetDimension].values.tolist()

      if self.maxTargetVariation(sourceDims, targetDim) != 1:
         raise Exception('Sources do not have enough variation for target. No solution found.') 

      lsCore = self.sourceList.copy()
      #'iterate through source and drop unnecessary ones one by one

      for i, v in reversed(list(enumerate(lsCore))):
        lsCoreTemp = lsCore.copy()
        if len(lsCoreTemp)> 1:
          lsCoreTemp.pop(i)
          if self.maxTargetVariation(self.generateGroupDimension(lsCoreTemp).values.tolist(), targetDim) == 1: 
            lsCore.pop(i)
    
      return lsCore

if __name__ == "__main__":      

    # ---------------The code from here is used for illustration of the class and could be translated into test cases----------------
    dictMapping = {'Indicator':['MP','MS','EP','ES','MP','MP'], 'Activity':['C','C','D','D','C','C'],'Measure':['P','S','P','S','P','P']}
    testdf = pd.DataFrame(dictMapping)

    print('\n Mapping tables')
    testMU = MappingUtility(testdf, ['Activity','Measure'], ['Indicator'])
    print(testMU.mappingDafaFrame)

    file_loc = "https://raw.githubusercontent.com/OECDSTD/SampleFilesForColab/main/test-mapping.csv"
    test2MU = MappingUtility.from_csv(file_loc,['DOMAIN', 'SUBJECT_S', 'COICOP', 'ACTIVITY', 'ADJUSTMENT_S','UNIT_MEASURE', 'BASE_PERIOD'], ['SUBJECT_T', 'ADJUSTMENT_T', 'UNIT', 'MEASURE'])
    print(test2MU.mappingDafaFrame)

    print('\n Dimension group example')
    gd = testMU.generateGroupDimension(['Activity','Measure'])
    print(gd)

    print('\n Single compatibility check between Activity and Indicator')
    print(testMU.maxTargetVariation(testMU.mappingDafaFrame['Activity'].values.tolist(), testMU.mappingDafaFrame['Indicator'].values.tolist()))

    print('\n Full compatibility check')
    print(testMU.analyseOneToOneCompatibility())
    print(test2MU.analyseOneToOneCompatibility())

    print('\n Generate the minimum set of source variables needed to describe a given target dimension')
    print('Indicator:', testMU.getMinimumCompatibilitySet('Indicator'))
    print('SUBJECT_T:', test2MU.getMinimumCompatibilitySet('SUBJECT_T'))



from copy import deepcopy
from typing import NamedTuple, Union
from dataclasses import dataclass
from enum import Enum
import re
import xml.etree.ElementTree as ET

NS = {"xsi": "http://www.w3.org/2001/XMLSchema-instance",
        "message": "http://www.sdmx.org/resources/sdmxml/schemas/v3_0/message",
        "str": "http://www.sdmx.org/resources/sdmxml/schemas/v3_0/structure",
        "com": "http://www.sdmx.org/resources/sdmxml/schemas/v3_0/common"
        }

INVALID = '!_INVALID_'


class MapType(Enum):
   OneToOne = 0
   OneToMany = 1
   ManyToOne = 2
   ManyToMany = 3

class SourceField(NamedTuple):
  isRegEx: bool
  value: str   


@dataclass
class Dataflow():
   urn: str
   isexternalreference: bool
   agencyID: str
   id: str
   version: str
   datastructure: str

   def __init__(self, elem: ET.Element) -> None:
      self.urn = elem.get('urn')
      if elem.get('isExternalReference') == 'false': 
         self.isexternalreference = False
      else:
         self.isexternalreference = True
      self.agencyID = elem.get('agencyID')
      self.id = elem.get('id')
      self.version = elem.get('version')
      self.datastructure = elem.find('str:Structure', namespaces=NS).text

   @property
   def fullid(self):
      return f'{self.agencyID}:{self.id}({self.version})'   
     


@dataclass
class DataStructure():
   urn: str
   isexternalreference: bool
   agencyID: str
   id: str
   version: str
   dimensions: list

   def __init__(self, elem: ET.Element) -> None:
      self.urn = elem.get('urn')
      if elem.get('isExternalReference') == 'false': 
         self.isexternalreference = False
      else:
         self.isexternalreference = True
      self.agencyID = elem.get('agencyID')
      self.id = elem.get('id')
      self.version = elem.get('version')

      self.dimensions = [] 
      for dimensions_element in elem.findall('*/str:DimensionList/str:Dimension', namespaces=NS):
         dimension = {k: dimensions_element.get(k) for k in ['id', 'position']}
         self.dimensions.append(deepcopy(dimension))
      for dimensions_element in elem.findall('*/str:DimensionList/str:TimeDimension', namespaces=NS):
         dimension = {k: dimensions_element.get(k) for k in ['id', 'position']}
         self.dimensions.append(deepcopy(dimension))


   def isdimension(self, dimension) -> bool:
      for d in self.dimensions:
         if d['id'] == dimension:
            return True

      return False
   
   def dim_list(self) -> list:
      return [d['id'] for d in self.dimensions]       


@dataclass
class StructureMap():
   source: str 
   target: str
   target_type: str = ''
   target_id: str = '' 
   
   def __init__(self, elem: ET.Element) -> None:
      self.source = elem.find('str:Source', namespaces=NS).text
      self.target = elem.find('str:Target', namespaces=NS).text
      p = re.compile(r'.*datastructure\.(.+)=(.+:.+\([0-9]+\.[0-9]+\))')
      m = p.match(self.target)      
      if m:
         self.target_type, self.target_id = m.groups()
     

@dataclass
class ComponentMap():
   sources: list
   targets: list
   type: MapType
   implicit: bool
   representation: str = ''
  
   def __init__(self, elem: ET.Element) -> None:
      sl = list(elem.findall('str:Source', namespaces=NS))
      self.sources = ['_S_'+e.text for e in sl]

      tl = list(elem.findall('str:Target', namespaces=NS))
      self.targets = [e.text for e in tl]

      rm = elem.findall('str:RepresentationMap', namespaces=NS)
      if rm:
            self.representation = rm[0].text
            self.implicit = False
      else:
            self.implicit = True

      if len(sl)==1:
            if len(tl)==1:
               self.type = MapType.OneToOne
            else:
               self.type = MapType.OneToMany
      else:
         if len(tl)==1:
            self.type = MapType.ManyToOne
         else:
            self.type = MapType.ManyToMany


@dataclass
class FixedValueMap():
   value: str
   target: str
  
   def __init__(self, elem: ET.Element) -> None:
      self.target = elem.find('str:Target', namespaces=NS).text
      self.value = elem.find('str:Value', namespaces=NS).text


@dataclass
class RepresentationMap():
   urn: str
   isexternalreference: bool
   mappings: list
   target_component_count: int

   def __init__(self, elem: ET.Element) -> None:
      self.urn = elem.get('urn')

      if elem.get('isExternalReference') == 'false': 
         self.isexternalreference = False
      else:
         self.isexternalreference = True

      self.target_component_count= len(elem.findall('str:TargetCodelist', namespaces=NS)) + len(elem.findall('str:TargetDataType', namespaces=NS))

      self.mappings = [] 
      for rm in elem.findall('str:RepresentationMapping', namespaces=NS):
         rmd = {'sourcevalues': [], 'targetvalues': []}
         for sv in rm.findall('str:SourceValue', namespaces=NS):
            if 'isRegEx' in sv.attrib.keys():
               rmd['sourcevalues'].append(SourceField(value=sv.text, isRegEx=bool(sv.attrib['isRegEx'])))
            else:   
               rmd['sourcevalues'].append(SourceField(value=sv.text, isRegEx=False))   
               
         for tv in rm.findall('str:TargetValue', namespaces=NS):
            rmd['targetvalues'].append(tv.text)
         padding = self.target_component_count - len(rmd['targetvalues'])
         for i in range(padding):
            rmd['targetvalues'].append(None)
         self.mappings.append(deepcopy(rmd))
      
   def _replace_matches(self, matching_groups: tuple, target_list: list[str]) -> list[str]:
      sp = re.compile(r'.*(\\[1-9]).*')
      for idx, val in enumerate(target_list):
         sm = sp.match(val)
         if sm: 
            for x in sm.groups():
               # index correction to allow end-user to use 1 based indexing for matching groups
               target_list[idx] = val.replace(x, matching_groups[int(x[1])-1], 1)
   
      return target_list

   def get_target_values_by_sourcelist(self, sourcelst: list) -> Union[list[str], None]:
      
         for m in self.mappings:
            pairs = zip(m['sourcevalues'], sourcelst)
            matched = True
            target = deepcopy(m['targetvalues'])
            # assumption: only one of the fields in source has RegEx
            for pair in pairs:
               if pair[0].isRegEx:
                  p = re.compile(pair[0].value)
                  matches = p.match(pair[1])
                  if matches:
                     target = self._replace_matches(matches.groups(), deepcopy(target))
                  else: 
                     matched = False
               else:
                  if pair[0].value != pair[1]:
                     matched = False
            if matched:
               return target
         if any([s!='' for s in sourcelst]):
            return [INVALID] * len(self.mappings[0]['targetvalues'])
         else:
            return [''] * len(self.mappings[0]['targetvalues'])   


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['mappings_utility']

package_data = \
{'': ['*']}

install_requires = \
['fastapi>=0.87.0,<0.88.0',
 'pandas>=1.5.0,<2.0.0',
 'requests>=2.28.1,<3.0.0',
 'uvicorn>=0.20.0,<0.21.0']

setup_kwargs = {
    'name': 'mappings-utility',
    'version': '0.1.6',
    'description': 'mapping utility tool for sdmx based partial mappings',
    'long_description': 'An SDMX mapping utility to generate partial key maps for referential metadata.\nAlthough not designed to do that, data mappings are also possible.\n\nThree methods are available:\n- *map_withFile* GET method (receiving the mapping source and mapping rules both as file references)\n- *map_withURN* GET method (receiving the mapping source  as file and the mapping rules as SDMX registry endpoint + mapping ID)\n- *map_json_withURN* POST method (receiving the mapping source as the body of the request in json - pandas dataframe dictionary style - and mapping rules as SDMX registry endpoint + mapping ID) (added in version 0.1.1)\n',
    'author': 'Gyorgy Gyomai',
    'author_email': 'gyorgy.gyomai@oecd.org',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



from typing import Literal, Union
import pandas as pd
import xml.etree.ElementTree as ET
import requests
from pathlib import Path
from copy import deepcopy
from pydantic import BaseModel
import logging


from mappings_utility.sdmx_structures import NS, INVALID, ComponentMap, DataStructure, Dataflow, FixedValueMap, MapType, RepresentationMap, StructureMap 


class KeyRange(BaseModel):
    __root__: dict[str, list[str]]


# Partial key mapping utility working with SDMX 3.0 structure-set and representation mapping objects
class SDMXMappingUtility():

   def __init__(self, mapping_tree: ET.ElementTree , source_file: Union[Path, None]  = None, json: Union[KeyRange, None] = None, sid_based: bool = False):
         self.source_file = source_file
         self.mapping_tree = mapping_tree
         self.json = json
         self._read_mapping()
         if source_file:
            self._read_source('csv', sid_based)
         elif json:
            self._read_source('json', sid_based)
      
   @classmethod
   def initialise_with_registry(cls, registry_url: str, source_file: Path, sid_based: bool = False):
      headers = {
         'accept': 'application/vnd.sdmx.structure+xml;version=3.0'
      }
      response = requests.get(registry_url, headers=headers)
      response.raise_for_status()
      return cls(ET.ElementTree(ET.fromstring(response.content)), source_file, sid_based=sid_based)

   @classmethod
   def initialise_with_jsonandregistry(cls, registry_url: str, json: KeyRange,  sid_based: bool = False):
      headers = {
         'accept': 'application/vnd.sdmx.structure+xml;version=3.0'
      }
      response = requests.get(registry_url, headers=headers)
      response.raise_for_status()
      return cls(ET.ElementTree(ET.fromstring(response.content)), json=json, sid_based=sid_based)

   @classmethod
   def initialise_with_file(cls, mapping_file: Path, source_file: Path, sid_based: bool = False):
      return cls(ET.parse(mapping_file), source_file, sid_based=sid_based)         
               

   def _read_mapping(self):
      logging.info('reading mapping objects...')
      try:
         root = self.mapping_tree.getroot()

         dataflow_elements = root.findall('*/str:Dataflows/str:Dataflow', namespaces=NS)
         dataflows = [Dataflow(e) for e in dataflow_elements]
         self.dataflows = dataflows

         datastructure_elements = root.findall('*/str:DataStructures/str:DataStructure', namespaces=NS)
         datastructures = [DataStructure(e) for e in datastructure_elements]
         self.datastructures = datastructures

         struct_map = root.find('*/str:StructureMaps/str:StructureMap', namespaces=NS)
         self.structure_map = StructureMap(struct_map)

         component_maps = list(struct_map.findall('str:ComponentMap', namespaces=NS))
         cms = [ComponentMap(e) for e in component_maps]
         self.component_maps = cms

         fixedvalue_maps = list(struct_map.findall('str:FixedValueMap', namespaces=NS))
         fvms = [FixedValueMap(e) for e in fixedvalue_maps]
         self.fixedvalue_maps = fvms

         representations = root.findall('*/str:RepresentationMaps/str:RepresentationMap', namespaces=NS)
         reps = [RepresentationMap(e) for e in representations]
         self.representation_maps = reps
      except Exception as err:
         logging.debug(str(err))
         raise err

   def _read_source(self, stype: Literal['csv', 'json'], sid_based: bool = False):
      logging.info('reading source file...')
      try:
         if stype == 'csv':
            df = pd.read_csv(self.source_file, dtype=str, na_filter=False)
         elif stype == 'json':
            df = pd.DataFrame.from_dict(self.json.dict()['__root__'], dtype=str)
         else:
            raise ValueError
      except Exception as err:
         logging.debug(str(err))
         raise err

      if sid_based:
            dsd = self.get_dsd_by_urn(self.structure_map.source)
            source_cols = dsd.dim_list()
            df[source_cols] = df.SID.str.split('.', expand=True)

      self.df_source_partial_keys = self._source_data_cosmetics(df)

   def _source_data_cosmetics(self, df):
         df.rename(columns = {cn: '_S_'+cn for cn in df.columns}, inplace = True)
         # Add missing columns
         for cm in self.component_maps:
            for source in cm.sources:
               if source not in df.columns:
                  df[source] = ''
         return df 

   def get_representation_by_urn(self, urn: str) -> Union[RepresentationMap, None]:
      for x in self.representation_maps:
         if x.urn == urn:
            return x
      return None

   # the urn to lookup could be that of a Dataflow or the DSD itself 
   def get_dsd_by_urn(self, urn: str) -> DataStructure:
      for df in self.dataflows:
         if df.urn == urn:
            for ds in self.datastructures:
               if ds.urn == df.datastructure:
                  return ds

      for dsd in self.datastructures:
         if dsd.urn == urn:
            return dsd

      return None                   

   # Generate mappings partial keys
   # includeSourceColumns: if true: both keys (from source and from target) will be in the dataframe
   #          false: only keys from target dataflow will be in the dataframe
   # includeAttributesMeasures: if true: all dimensions, attributes, and measures of the target dataflow will be in the dataframe
   #          false: only dimensions from target dataflow will be in the dataframe
   # nulledFixedTargets: if true: the targets with fixed representation will not be inserted (the expected behaviour when partial keys are mapped 
   #          for attribute and referential metadata attachment), false: for full data mappings
   # writeTargetStructInfo: if true: SDMX csv 3.0 style structural and action information will be writtern into the first columns of the generated file
   # dropInvalid: if true: non-mapped and source records containing invalid members will be dropped (in cases when the mapping acts as a filter)
   # sid_based: if true read SID as co-ordinates of the source dimensions in the correct order and deliver SID according to target
   

   def generate_mappings_partial_keys(self, includeSourceColumns=False, includeAttributesMeasures=False, nulledFixedTargets=False, writeTargetStructInfo=False, dropInvalid=False, sid_based=False):
      
         df = self.df_source_partial_keys
         
         # TODO revise the structure components to write (is source needed?, controlling the Action - with an Enum?, Generate SID?)
         al = ['STRUCTURE', 'STRUCTURE_ID', 'ACTION']
         if writeTargetStructInfo:
            df[al] = df.apply(lambda row: [self.structure_map.target_type.lower(), self.structure_map.target_id, 'A'], axis=1, result_type='expand')

         # ComponentMaps
         for cm in self.component_maps:
            if cm.type in [MapType.OneToOne] and cm.implicit:
               # copy identical with target component-name
               logging.info(f'copying source to target {cm.sources} -> {cm.targets}')
               df[cm.targets[0]] = df[cm.sources[0]]
            else:
               # only those cases can be mapped safely where there are valid inputs in all source columns
               logging.info(f'applying representation map {cm.sources} -> {cm.targets}')
               rm = self.get_representation_by_urn(cm.representation)
               df[cm.targets] = df.apply(lambda x: rm.get_target_values_by_sourcelist([x[c] for c in cm.sources]), axis=1, result_type='expand')

         # Add target missing columns mapped as fixed value
         for fvm in self.fixedvalue_maps:
            if not nulledFixedTargets:
               df[fvm.target] = fvm.value
            else:
               df[fvm.target] = ''

         if sid_based:
            dsd = self.get_dsd_by_urn(self.structure_map.target)
            df['SID'] = df.apply(lambda x: '.'.join([x[c] for c in dsd.dim_list()]), axis=1)

         # Remove source columns which are in the mappings and some technical columns      
         if not includeSourceColumns:
            # Get mapped source columns 
            source_columns = []
            for cm in self.component_maps:
               source_columns += cm.sources
            # Remove technical columns if they exist
            source_columns += ['_S_STRUCTURE', '_S_STRUCTURE_ID', '_S_ACTION', '_S_SID']
            # Create a filter from the "source_columns" that exists in df
            columns_todrop = df.filter(source_columns).columns
            # Drop the all columns in this list: columns_todrop
            df.drop(columns_todrop, axis=1, inplace=True)

         # Remove columns, which are not in the dimension list of the target dataflow
         # such as attributes and measures
         if not includeAttributesMeasures:
            dsd = self.get_dsd_by_urn(self.structure_map.target)
            if sid_based:
               columns_to_drop = [column for column in df.columns if (not column =='SID' and not column.startswith("_S_") and not column in al)]
            else: 
               columns_to_drop = [column for column in df.columns if (not dsd.isdimension(column) and not column.startswith("_S_") and not column in al)]
            df.drop(columns_to_drop, axis=1, inplace=True)

         if dropInvalid:
            df = df[df.ne(INVALID).all(axis=1)]   

         return deepcopy(df)